# Actions that can be performed by the system in order to perform an
# upgrade. The manifest is a list of these.

import os.path
import shutil
import zlib

import mutable.v2.phase2.gui as gui #@UnresolvedImport

class Action(object):

    def affected(self):
        """
        Returns a list of filesystem paths that will be affected by this
        operation.
        """

        assert False

    def perform(self, p2):
        """
        Performs the operation in question. p2 is the p2 object. 
        """

        assert False
        
    def __repr__(self):
        """
        All actions must define a __repr__ method.
        """

        assert False

def clean(fn):
    """
    Unconditionally removes the given name.
    """

    if os.path.islink(fn):
        os.unlink(fn)
        return
        
    if not os.path.exists(fn):
        return

    if os.path.isdir(fn):
        shutil.rmtree(fn)
    else:
        os.unlink(fn)

    
def safe_rename(old, new):
    """
    This renames old to new. It tries to do this atomically, but may need
    to recursively delete new if it is a directory.
    """

    if os.path.islink(new):
        clean(new)
        
    try:
        os.rename(old, new)
    except OSError:
        clean(new)        
        os.rename(old, new)


def set_perms(fn, perms):
    """
    Sets the permissions on a file. The permissions are computed by
    preserve().
    """

    mode, uid, gid, atime, mtime = perms

    os.lchown(fn, uid, gid)
    
    if not os.path.islink(fn):
        os.chmod(fn, mode)
        os.utime(fn, (atime, mtime))
        
class Unpack(Action):
    """
    Unpacks a filename from the archive.
    """

    def __init__(self, filename, perms):
        super(Unpack, self).__init__()

        self.filename = filename
        self.perms = perms

    def affected(self):
        return [ self.filename ]

    def perform(self, p2):
        tmp = p2.root + self.filename + ".stunew"
        clean(tmp)

        # f = file(tmp, "w")
        # f.write(p2.stu.read("data" + self.filename))
        # f.close()
        p2.stu.read_into("data" + self.filename, tmp)

        
        set_perms(tmp, self.perms)
        safe_rename(tmp, p2.root + self.filename)

    def __repr__(self):
        return "Unpack(%r, %r)" % (self.filename, self.perms)

    def priority(self):
        return (1, self.filename)

class BSPatch(Action):
    """
    Uses BSDiff to patch a file.
    """

    def __init__(self, oldfn, newfn, perms, oldadler, newadler):
        super(BSPatch, self).__init__()

        self.oldfn = oldfn
        self.newfn = newfn
        self.perms = perms
        self.oldadler = oldadler
        self.newadler = newadler

    def affected(self):
        return [ self.filename ]

    def perform(self, p2):
        if self.oldfn == "/dev/null":
            old = "/dev/null"
        else:        
            old = p2.root + self.oldfn
        
        new = p2.root + self.newfn
        tmp = "/bsdiff.new"
        
        crc = zlib.adler32("")

        f = file(old, "rb")
        while True:
            data = f.read(65536)
            if not data:
                break
            crc = zlib.adler32(data, crc)        
        f.close()

        if crc != self.oldadler:
            if crc != self.newadler and p2.inhouse:
                raise Exception("Unknown file checksum.")
            return
        
        clean(tmp)

        patch = "/tmp/bsdiff.patch"
        p2.stu.read_into("data" + self.newfn + ".patch", patch)

        # Fugly code, since we can't fork to use subprocess or os.spawn, and we don't 
        # want to have to shell-quote to use os.system.
        os.symlink(old, "/tmp/bsdiff.old")       
        os.system('/usr/bin/bspatch /tmp/bsdiff.old %s %s' % (tmp, patch))
        os.unlink("/tmp/bsdiff.old")
        
        set_perms(tmp, self.perms)
        safe_rename(tmp, new)

    def __repr__(self):
        return "BSPatch(%r, %r, %r, %r, %r)" % (self.oldfn, self.newfn, self.perms, self.oldadler, self.newadler)

    def priority(self):
        return (1, self.newfn)
    
class Mkdir(Action):

    def __init__(self, filename, perms):
        super(Mkdir, self).__init__()

        self.filename = filename
        self.perms = perms

    def affected(self):
        return [ self.filename ]

    def perform(self, p2):
        if os.path.isdir(p2.root + self.filename):
            set_perms(p2.root + self.filename, self.perms)
        else:
            fn = p2.root + self.filename
            os.mkdir(fn, 0777)
            set_perms(fn, self.perms)
            
    def __repr__(self):
        return "Mkdir(%r, %r)" % (self.filename, self.perms)
                     

    def priority(self):
        return (0, self.filename)
    
class Symlink(Action):

    def __init__(self, filename, target, perms):
        super(Symlink, self).__init__()

        self.filename = filename
        self.target = target
        self.perms = perms

    def affected(self):
        return [ self.filename ]

    def perform(self, p2):
        clean(p2.root + self.filename)
        os.symlink(self.target, p2.root + self.filename)
        set_perms(p2.root + self.filename, self.perms)

    def __repr__(self):
        return "Symlink(%r, %r, %r)" % (self.filename, self.target, self.perms)
        
    def priority(self):
        return (2, self.filename)

class Mknod(Action):

    def __init__(self, filename, type, device, perms):
        super(Mknod, self).__init__()

        self.filename = filename
        self.type = type
        self.device = device
        self.perms = perms

    def affected(self):
        return [ self.filename ]

    def perform(self, p2):
        tmp = p2.root + self.filename + ".stunew"
        clean(tmp)
        
        os.mknod(tmp, self.type, self.device)
        set_perms(tmp, self.perms)
        safe_rename(tmp, p2.root + self.filename)

    def __repr__(self):
        return "Mknod(%r, %r, %r, %r)" % (self.filename, self.type, self.device, self.perms)

    def priority(self):
        return (1, self.filename)
    
class Unlink(Action):

    def __init__(self, filename):
        super(Unlink, self).__init__()

        self.filename = filename
        
    def affected(self):
        return [ self.filename ]

    def perform(self, p2):
        clean(p2.root + self.filename)

    def __repr__(self):
        return "Unlink(%r)" % (self.filename,)

    def priority(self):
        return (3, self.filename)


def perform(p2):
    """
    This actually performs the actions listed in the manifest of the
    given p2 object.
    """

    namespace = dict(Unpack=Unpack,
                     Mkdir=Mkdir,
                     Symlink=Symlink,
                     Mknod=Mknod,
                     Unlink=Unlink,
                     BSPatch=BSPatch)

    p2.stu.read_into("meta/manifest", "/tmp/manifest.txt")

    manifest_len = 0
    for l in file("/tmp/manifest.txt"):
        manifest_len += 1
    
    i = 0
    for l in file("/tmp/manifest.txt"):
        if not l:
            continue

        a = eval(l, namespace)
        print "Performing", a 
        
        gui.progress(i, manifest_len)
        i += 1

        a.perform(p2)

    os.unlink("/tmp/manifest.txt")
